@@ -103,13 +103,6 @@ static void      menu_tree_resolve_files        (MenuTree       *tree,
 static void      menu_tree_force_recanonicalize (MenuTree       *tree);
 static void      menu_tree_invoke_monitors      (MenuTree       *tree);
      
-static void menu_tree_add_menu_file_monitor     (MenuTree   *tree,
-						 const char *path,
-						 gboolean    existent);
-static void menu_tree_remove_menu_file_monitors (MenuTree   *tree);
-
-
-
 /*
  * The idea is that we cache the menu tree for either a given
  * menu basename or an absolute menu path.
@@ -174,6 +167,14 @@ menu_tree_remove_from_cache (MenuTree *tree)
     }
 }
 
+typedef enum
+{
+  MENU_FILE_MONITOR_INVALID = 0,
+  MENU_FILE_MONITOR_FILE,
+  MENU_FILE_MONITOR_NONEXISTENT_FILE,
+  MENU_FILE_MONITOR_DIRECTORY
+} MenuFileMonitorType;
+
 static void
 handle_nonexistent_menu_file_changed (GnomeVFSMonitorHandle    *handle,
 				      const char               *monitor_uri,
@@ -184,9 +185,9 @@ handle_nonexistent_menu_file_changed (GnomeVFSMonitorHandle    *handle,
   if (event == GNOME_VFS_MONITOR_EVENT_CHANGED ||
       event == GNOME_VFS_MONITOR_EVENT_CREATED)
     {
-      menu_verbose ("File \"%s\" %s, marking tree for recanonicalization\n",
+      menu_verbose ("\"%s\" %s, marking tree for recanonicalization\n",
                     info_uri,
-                    event == GNOME_VFS_MONITOR_EVENT_CREATED ? ("created") : ("changed"));
+                    event == GNOME_VFS_MONITOR_EVENT_CREATED ? "created" : "changed");
 
       menu_tree_force_recanonicalize (tree);
       menu_tree_invoke_monitors (tree);
@@ -200,56 +201,92 @@ handle_menu_file_changed (GnomeVFSMonitorHandle    *handle,
                           GnomeVFSMonitorEventType  event,
                           MenuTree                 *tree)
 {
-  if (event == GNOME_VFS_MONITOR_EVENT_DELETED)
-    {
-      menu_verbose ("File \"%s\" deleted, marking tree for recanicalization\n",
-                    info_uri);
+  if (event != GNOME_VFS_MONITOR_EVENT_DELETED &&
+      event != GNOME_VFS_MONITOR_EVENT_CHANGED &&
+      event != GNOME_VFS_MONITOR_EVENT_CREATED)
+    return;
 
-      menu_tree_force_recanonicalize (tree);
-      menu_tree_invoke_monitors (tree);
-    }
-  else if (event == GNOME_VFS_MONITOR_EVENT_CHANGED ||
-	   event == GNOME_VFS_MONITOR_EVENT_CREATED)
-    {
-      menu_verbose ("File \"%s\" %s, marking layout for reload\n",
-                    monitor_uri,
-                    event == GNOME_VFS_MONITOR_EVENT_CREATED ? ("created") : ("changed"));
+  menu_verbose ("\"%s\" %s, marking tree for recanicalization\n",
+		info_uri,
+		event == GNOME_VFS_MONITOR_EVENT_CREATED ? "created" :
+		event == GNOME_VFS_MONITOR_EVENT_CHANGED ? "changed" : "deleted");
 
-      menu_tree_force_reload (tree);
-      menu_tree_invoke_monitors (tree);
-    }
+  menu_tree_force_recanonicalize (tree);
+  menu_tree_invoke_monitors (tree);
 }
 
 static void
-menu_tree_add_menu_file_monitor (MenuTree   *tree,
-                                 const char *path,
-				 gboolean    existent)
+handle_menu_file_directory_changed (GnomeVFSMonitorHandle    *handle,
+				    const char               *monitor_uri,
+				    const char               *info_uri,
+				    GnomeVFSMonitorEventType  event,
+				    MenuTree                 *tree)
+{
+  if (event != GNOME_VFS_MONITOR_EVENT_DELETED &&
+      event != GNOME_VFS_MONITOR_EVENT_CHANGED &&
+      event != GNOME_VFS_MONITOR_EVENT_CREATED)
+    return;
+
+  if (!g_str_has_suffix (info_uri, ".menu"))
+    return;
+
+  menu_verbose ("\"%s\" %s, marking tree for recanicalization\n",
+		info_uri,
+		event == GNOME_VFS_MONITOR_EVENT_CREATED ? "created" :
+		event == GNOME_VFS_MONITOR_EVENT_CHANGED ? "changed" : "deleted");
+
+  menu_tree_force_recanonicalize (tree);
+  menu_tree_invoke_monitors (tree);
+}
+
+static void
+menu_tree_add_menu_file_monitor (MenuTree            *tree,
+                                 const char          *path,
+				 MenuFileMonitorType  type)
 {
   GnomeVFSMonitorHandle *handle;
   GnomeVFSResult         result;
   char                  *uri;
 
-  menu_verbose ("Adding a menu file monitor for %sexistent \"%s\"\n",
-		existent ? "" : "non", path);
-
   uri = gnome_vfs_get_uri_from_local_path (path);
 
   handle = NULL;
-  if (existent)
+
+  switch (type)
     {
+    case MENU_FILE_MONITOR_FILE:
+      menu_verbose ("Adding a menu file monitor for \"%s\"\n", path);
+
       result = gnome_vfs_monitor_add (&handle,
 				      uri,
 				      GNOME_VFS_MONITOR_FILE,
 				      (GnomeVFSMonitorCallback) handle_menu_file_changed,
 				      tree);
-    }
-  else
-    {
+      break;
+
+    case MENU_FILE_MONITOR_NONEXISTENT_FILE:
+      menu_verbose ("Adding a menu file monitor for non-existent \"%s\"\n", path);
+
       result = gnome_vfs_monitor_add (&handle,
 				      uri,
 				      GNOME_VFS_MONITOR_FILE,
 				      (GnomeVFSMonitorCallback) handle_nonexistent_menu_file_changed,
 				      tree);
+      break;
+
+    case MENU_FILE_MONITOR_DIRECTORY:
+      menu_verbose ("Adding a menu directory monitor for \"%s\"\n", path);
+
+      result = gnome_vfs_monitor_add (&handle,
+				      uri,
+				      GNOME_VFS_MONITOR_DIRECTORY,
+				      (GnomeVFSMonitorCallback) handle_menu_file_directory_changed,
+				      tree);
+      break;
+
+    default:
+      g_assert_not_reached ();
+      break;
     }
 
   if (result == GNOME_VFS_OK)
@@ -259,7 +296,7 @@ menu_tree_add_menu_file_monitor (MenuTree   *tree,
   else
     {
       g_assert (handle == NULL);
-      menu_verbose ("Failed to add file monitor to for %s: %s\n",
+      menu_verbose ("Failed to add monitor for %s: %s\n",
                     path, gnome_vfs_result_to_string (result));
     }
 
@@ -340,11 +377,15 @@ canonicalize_basename_with_config_dir (MenuTree   *tree,
   if (tree->canonical_path)
     {
       tree->canonical = TRUE;
-      menu_tree_add_menu_file_monitor (tree, tree->canonical_path, TRUE);
+      menu_tree_add_menu_file_monitor (tree,
+				       tree->canonical_path,
+				       MENU_FILE_MONITOR_FILE);
     }
   else
     {
-      menu_tree_add_menu_file_monitor (tree, path, FALSE);
+      menu_tree_add_menu_file_monitor (tree,
+				       path,
+				       MENU_FILE_MONITOR_NONEXISTENT_FILE);
     }
 
   g_free (path);
@@ -409,7 +450,7 @@ menu_tree_canonicalize_path (MenuTree *tree)
           menu_tree_remove_menu_file_monitors (tree);
           menu_tree_add_menu_file_monitor (tree,
                                            tree->canonical_path,
-					   TRUE);
+					   MENU_FILE_MONITOR_FILE);
 
           tree->canonical = TRUE;
 
@@ -483,11 +524,15 @@ menu_tree_new (MenuTreeType  type,
       if (tree->canonical)
 	{
 	  tree->canonical_path = g_strdup (menu_file);
-	  menu_tree_add_menu_file_monitor (tree, tree->canonical_path, TRUE);
+	  menu_tree_add_menu_file_monitor (tree,
+					   tree->canonical_path,
+					   MENU_FILE_MONITOR_FILE);
 	}
       else
 	{
-	  menu_tree_add_menu_file_monitor (tree, tree->absolute_path, FALSE);
+	  menu_tree_add_menu_file_monitor (tree,
+					   tree->absolute_path,
+					   MENU_FILE_MONITOR_NONEXISTENT_FILE);
 	}
     }
 
@@ -1082,10 +1127,11 @@ merge_resolved_children (MenuTree       *tree,
     }
 }
 
-static void
+static gboolean
 load_merge_file (MenuTree       *tree,
                  const char     *filename,
                  gboolean        is_canonical,
+		 gboolean        add_monitor,
                  MenuLayoutNode *where)
 {
   MenuLayoutNode *to_merge;
@@ -1097,9 +1143,14 @@ load_merge_file (MenuTree       *tree,
       canonical = freeme = menu_canonicalize_file_name (filename, FALSE);
       if (canonical == NULL)
         {
+	  if (add_monitor)
+	    menu_tree_add_menu_file_monitor (tree,
+					     filename,
+					     MENU_FILE_MONITOR_NONEXISTENT_FILE);
+
           menu_verbose ("Failed to canonicalize merge file path \"%s\": %s\n",
                         filename, g_strerror (errno));
-          return;
+          return FALSE;
         }
     }
   else
@@ -1114,10 +1165,13 @@ load_merge_file (MenuTree       *tree,
     {
       menu_verbose ("No menu for file \"%s\" found when merging\n",
                     canonical);
-      return;
+      return FALSE;
     }
 
-  menu_tree_add_menu_file_monitor (tree, canonical, TRUE);
+  if (add_monitor)
+    menu_tree_add_menu_file_monitor (tree,
+				     canonical,
+				     MENU_FILE_MONITOR_FILE);
 
   merge_resolved_children (tree, where, to_merge);
 
@@ -1125,6 +1179,121 @@ load_merge_file (MenuTree       *tree,
 
   if (freeme)
     g_free (freeme);
+
+  return TRUE;
+}
+
+static gboolean
+load_merge_file_with_config_dir (MenuTree       *tree,
+				 const char     *menu_file,
+				 const char     *config_dir,
+				 MenuLayoutNode *where)
+{
+  char     *merge_file;
+  gboolean  loaded;
+
+  loaded = FALSE;
+
+  merge_file = g_build_filename (config_dir, "menus", menu_file, NULL);
+
+  if (load_merge_file (tree, merge_file, FALSE, TRUE, where))
+    loaded = TRUE;
+
+  g_free (merge_file);
+
+  return loaded;
+}
+
+static gboolean
+compare_basedir_to_config_dir (const char *canonical_basedir,
+			       const char *config_dir)
+{
+  char     *dirname;
+  char     *canonical_menus_dir;
+  gboolean  retval;
+
+  menu_verbose ("Checking to see if basedir '%s' is in '%s'\n",
+		canonical_basedir, config_dir);
+
+  dirname = g_build_filename (config_dir, "menus", NULL);
+
+  retval = FALSE;
+
+  canonical_menus_dir = menu_canonicalize_file_name (dirname, FALSE);
+  if (canonical_menus_dir != NULL &&
+      strcmp (canonical_basedir, canonical_menus_dir) == 0)
+    {
+      retval = TRUE;
+    }
+
+  g_free (canonical_menus_dir);
+  g_free (dirname);
+
+  return retval;
+}
+
+static gboolean
+load_parent_merge_file (MenuTree       *tree,
+			MenuLayoutNode *layout)
+{
+  MenuLayoutNode     *root;
+  const char         *basedir;
+  const char         *menu_name;
+  char               *canonical_basedir;
+  char               *menu_file;
+  gboolean            found_basedir;
+  const char * const *system_config_dirs;
+  int                 i;
+
+  root = menu_layout_node_get_root (layout);
+
+  basedir   = menu_layout_node_root_get_basedir (root);
+  menu_name = menu_layout_node_root_get_name (root);
+
+  canonical_basedir = menu_canonicalize_file_name (basedir, FALSE);
+  if (canonical_basedir == NULL)
+    {
+      menu_verbose ("Menu basedir '%s' no longer exists, not merging parent\n",
+		    basedir);
+      return FALSE;
+    }
+
+  menu_file = g_strconcat (menu_name, ".menu", NULL);
+
+  found_basedir = compare_basedir_to_config_dir (canonical_basedir,
+						 g_get_user_config_dir ());
+
+  system_config_dirs = g_get_system_config_dirs ();
+
+  i = 0;
+  while (system_config_dirs[i] != NULL)
+    {
+      if (!found_basedir)
+	{
+	  found_basedir = compare_basedir_to_config_dir (canonical_basedir,
+							 system_config_dirs[i]);
+	}
+      else
+	{
+	  menu_verbose ("Looking for parent menu file '%s' in '%s'\n",
+			menu_file, system_config_dirs[i]);
+
+	  if (load_merge_file_with_config_dir (tree,
+					       menu_file,
+					       system_config_dirs[i],
+					       layout))
+	    {
+	      break;
+	    }
+	}
+
+      ++i;
+    }
+
+  g_free (menu_file);
+  g_free (canonical_basedir);
+
+  return found_basedir;
 }
 
 static void
@@ -1137,6 +1306,10 @@ load_merge_dir (MenuTree       *tree,
 
   menu_verbose ("Loading merge dir \"%s\"\n", dirname);
 
+  menu_tree_add_menu_file_monitor (tree,
+				   dirname,
+				   MENU_FILE_MONITOR_DIRECTORY);
+
   if ((dir = g_dir_open (dirname, 0, NULL)) == NULL)
     return;
 
@@ -1148,7 +1321,7 @@ load_merge_dir (MenuTree       *tree,
 
           full_path = g_build_filename (dirname, menu_file, NULL);
 
-          load_merge_file (tree, full_path, TRUE, where);
+          load_merge_file (tree, full_path, TRUE, FALSE, where);
 
           g_free (full_path);
         }
@@ -1178,6 +1351,12 @@ resolve_merge_file (MenuTree         *tree,
 {
   char *filename;
 
+  if (menu_layout_node_merge_file_get_type (layout) == MENU_MERGE_FILE_TYPE_PARENT)
+    {
+      if (load_parent_merge_file (tree, layout))
+        return;
+    }
+
   filename = menu_layout_node_get_content_as_path (layout);
   if (filename == NULL)
     {
@@ -1185,7 +1364,7 @@ resolve_merge_file (MenuTree         *tree,
     }
   else
     {
-      load_merge_file (tree, filename, FALSE, layout);
+      load_merge_file (tree, filename, FALSE, TRUE, layout);
 
       g_free (filename);
     }
@@ -1216,9 +1395,9 @@ resolve_merge_dir (MenuTree       *tree,
   menu_layout_node_unlink (layout);
 }
 
-static void
+static MenuLayoutNode *
 add_app_dir (MenuTree       *tree,
-             MenuLayoutNode *layout,
+             MenuLayoutNode *before,
              const char     *data_dir)
 {
   MenuLayoutNode *tmp;
@@ -1227,41 +1406,48 @@ add_app_dir (MenuTree       *tree,
   tmp = menu_layout_node_new (MENU_LAYOUT_NODE_APP_DIR);
   dirname = g_build_filename (data_dir, "applications", NULL);
   menu_layout_node_set_content (tmp, dirname);
-  menu_layout_node_insert_before (layout, tmp);
-  menu_layout_node_unref (tmp);
+  menu_layout_node_insert_before (before, tmp);
+  menu_layout_node_unref (before);
 
   menu_verbose ("Adding <AppDir>%s</AppDir> in <DefaultAppDirs/>\n",
                 dirname);
 
   g_free (dirname);
+
+  return tmp;
 }
 
 static void
 resolve_default_app_dirs (MenuTree       *tree,
                           MenuLayoutNode *layout)
 {
+  MenuLayoutNode     *before;
   const char * const *system_data_dirs;
-  int                i;
-
-  add_app_dir (tree, layout, g_get_user_data_dir ());
+  int                 i;
 
   system_data_dirs = g_get_system_data_dirs ();
 
+  before = add_app_dir (tree,
+			menu_layout_node_ref (layout),
+			g_get_user_data_dir ());
+
   i = 0;
   while (system_data_dirs[i] != NULL)
     {
-      add_app_dir (tree, layout, system_data_dirs[i]);
+      before = add_app_dir (tree, before, system_data_dirs[i]);
 
       ++i;
     }
 
+  menu_layout_node_unref (before);
+
   /* remove the now-replaced node */
   menu_layout_node_unlink (layout);
 }
 
-static void
+static MenuLayoutNode *
 add_directory_dir (MenuTree       *tree,
-                   MenuLayoutNode *layout,
+                   MenuLayoutNode *before,
                    const char     *data_dir)
 {
   MenuLayoutNode *tmp;
@@ -1270,34 +1456,41 @@ add_directory_dir (MenuTree       *tree,
   tmp = menu_layout_node_new (MENU_LAYOUT_NODE_DIRECTORY_DIR);
   dirname = g_build_filename (data_dir, "desktop-directories", NULL);
   menu_layout_node_set_content (tmp, dirname);
-  menu_layout_node_insert_before (layout, tmp);
-  menu_layout_node_unref (tmp);
+  menu_layout_node_insert_before (before, tmp);
+  menu_layout_node_unref (before);
 
   menu_verbose ("Adding <DirectoryDir>%s</DirectoryDir> in <DefaultDirectoryDirs/>\n",
                 dirname);
 
   g_free (dirname);
+
+  return tmp;
 }
 
 static void
 resolve_default_directory_dirs (MenuTree       *tree,
                                 MenuLayoutNode *layout)
 {
+  MenuLayoutNode     *before;
   const char * const *system_data_dirs;
-  int          i;
-
-  add_directory_dir (tree, layout, g_get_user_data_dir ());
+  int                 i;
 
   system_data_dirs = g_get_system_data_dirs ();
 
+  before = add_directory_dir (tree,
+			      menu_layout_node_ref (layout),
+			      g_get_user_data_dir ());
+
   i = 0;
   while (system_data_dirs[i] != NULL)
     {
-      add_directory_dir (tree, layout, system_data_dirs[i]);
+      before = add_directory_dir (tree, before, system_data_dirs[i]);
 
       ++i;
     }
 
+  menu_layout_node_unref (before);
+
   /* remove the now-replaced node */
   menu_layout_node_unlink (layout);
 }
@@ -1317,24 +1510,25 @@ resolve_default_merge_dirs (MenuTree       *tree,
 
   merge_name = g_strconcat (menu_name, "-merged", NULL);
 
-  load_merge_dir_with_config_dir (tree,
-                                  g_get_user_config_dir (),
-                                  merge_name,
-                                  layout);
-
   system_config_dirs = g_get_system_config_dirs ();
 
+  /* Merge in reverse order */
   i = 0;
-  while (system_config_dirs[i] != NULL)
+  while (system_config_dirs[i] != NULL) i++;
+  while (i > 0)
     {
+      i--;
       load_merge_dir_with_config_dir (tree,
                                       system_config_dirs[i],
                                       merge_name,
                                       layout);
-
-      ++i;
     }
 
+  load_merge_dir_with_config_dir (tree,
+                                  g_get_user_config_dir (),
+                                  merge_name,
+                                  layout);
+
   g_free (merge_name);
 
   /* remove the now-replaced node */
@@ -1361,6 +1555,7 @@ add_filename_include (const char     *desktop_file_id,
 static gboolean
 add_menu_for_legacy_dir (MenuLayoutNode *parent,
                          const char     *legacy_dir,
+                	 const char     *relative_path,
                          const char     *legacy_prefix,
                          const char     *menu_name)
 {
@@ -1382,14 +1577,15 @@ add_menu_for_legacy_dir (MenuLayoutNode *parent,
   entry_directory_unref (ed);
 
   menu_added = FALSE;
-  if (desktop_entry_set_get_count (desktop_entries) > 0)
+  if (desktop_entry_set_get_count (desktop_entries) > 0 || subdirs)
     {
       MenuLayoutNode *menu;
-      MenuLayoutNode *include;
       MenuLayoutNode *node;
       GString        *subdir_path;
+      GString        *subdir_relative;
       GSList         *tmp;
       int             legacy_dir_len;
+      int             relative_path_len;
 
       menu = menu_layout_node_new (MENU_LAYOUT_NODE_MENU);
       menu_layout_node_append_child (parent, menu);
@@ -1404,20 +1600,41 @@ add_menu_for_legacy_dir (MenuLayoutNode *parent,
       menu_layout_node_unref (node);
 
       node = menu_layout_node_new (MENU_LAYOUT_NODE_DIRECTORY);
-      menu_layout_node_set_content (node, ".directory");
+      if (relative_path != NULL)
+        {
+          char *directory_entry_path;
+
+          directory_entry_path = g_strdup_printf ("%s/.directory", relative_path);
+          menu_layout_node_set_content (node, directory_entry_path);
+          g_free (directory_entry_path);
+        }
+      else
+        {
+          menu_layout_node_set_content (node, ".directory");
+        }
       menu_layout_node_append_child (menu, node);
       menu_layout_node_unref (node);
 
-      include = menu_layout_node_new (MENU_LAYOUT_NODE_INCLUDE);
-      menu_layout_node_append_child (menu, include);
+      if (desktop_entry_set_get_count (desktop_entries) > 0)
+	{
+	  MenuLayoutNode *include;
 
-      desktop_entry_set_foreach (desktop_entries,
-                                 (DesktopEntrySetForeachFunc) add_filename_include,
-                                 include);
+	  include = menu_layout_node_new (MENU_LAYOUT_NODE_INCLUDE);
+	  menu_layout_node_append_child (menu, include);
+
+	  desktop_entry_set_foreach (desktop_entries,
+				     (DesktopEntrySetForeachFunc) add_filename_include,
+				     include);
+
+	  menu_layout_node_unref (include);
+	}
 
       subdir_path = g_string_new (legacy_dir);
       legacy_dir_len = strlen (legacy_dir);
 
+      subdir_relative = g_string_new (relative_path);
+      relative_path_len = relative_path ? strlen (relative_path) : 0;
+
       tmp = subdirs;
       while (tmp != NULL)
         {
@@ -1426,11 +1643,19 @@ add_menu_for_legacy_dir (MenuLayoutNode *parent,
           g_string_append_c (subdir_path, '/');
           g_string_append (subdir_path, subdir);
 
+	  if (relative_path_len)
+	    {
+	      g_string_append_c (subdir_relative, '/');
+	    }
+          g_string_append (subdir_relative, subdir);
+
           add_menu_for_legacy_dir (menu,
                                    subdir_path->str,
+				   subdir_relative->str,
                                    legacy_prefix,
                                    subdir);
 
+          g_string_truncate (subdir_relative, relative_path_len);
           g_string_truncate (subdir_path, legacy_dir_len);
 
           tmp = tmp->next;
@@ -1438,7 +1663,6 @@ add_menu_for_legacy_dir (MenuLayoutNode *parent,
 
       g_string_free (subdir_path, TRUE);
 
-      menu_layout_node_unref (include);
       menu_layout_node_unref (menu);
     }
 
@@ -1464,6 +1688,7 @@ resolve_legacy_dir (MenuTree       *tree,
 
   if (add_menu_for_legacy_dir (to_merge,
                                menu_layout_node_get_content (legacy),
+			       NULL,
                                menu_layout_node_legacy_dir_get_prefix (legacy),
                                menu_layout_node_menu_get_name (menu)))
     {
@@ -1473,9 +1698,9 @@ resolve_legacy_dir (MenuTree       *tree,
   menu_layout_node_unref (to_merge);
 }
 
-static void
+static MenuLayoutNode *
 add_legacy_dir (MenuTree       *tree,
-                MenuLayoutNode *layout,
+                MenuLayoutNode *before,
                 const char     *data_dir)
 {
   MenuLayoutNode *legacy;
@@ -1486,37 +1711,43 @@ add_legacy_dir (MenuTree       *tree,
   legacy = menu_layout_node_new (MENU_LAYOUT_NODE_LEGACY_DIR);
   menu_layout_node_set_content (legacy, dirname);
   menu_layout_node_legacy_dir_set_prefix (legacy, "kde");
-  menu_layout_node_insert_before (layout, legacy);
+  menu_layout_node_insert_before (before, legacy);
+  menu_layout_node_unref (before);
 
   menu_verbose ("Adding <LegacyDir>%s</LegacyDir> in <KDELegacyDirs/>\n",
                 dirname);
 
   resolve_legacy_dir (tree, legacy);
 
-  menu_layout_node_unref (legacy);
-
   g_free (dirname);
+
+  return legacy;
 }
 
 static void
 resolve_kde_legacy_dirs (MenuTree       *tree,
                          MenuLayoutNode *layout)
 {
+  MenuLayoutNode     *before;
   const char * const *system_data_dirs;
   int                 i;
 
-  add_legacy_dir (tree, layout, g_get_user_data_dir ());
-
   system_data_dirs = g_get_system_data_dirs ();
 
+  before = add_legacy_dir (tree,
+			   menu_layout_node_ref (layout),
+			   g_get_user_data_dir ());
+
   i = 0;
   while (system_data_dirs[i] != NULL)
     {
-      add_legacy_dir (tree, layout, system_data_dirs[i]);
+      before = add_legacy_dir (tree, before, system_data_dirs[i]);
 
       ++i;
     }
 
+  menu_layout_node_unref (before);
+
   /* remove the now-replaced node */
   menu_layout_node_unlink (layout);
 }
@@ -2108,6 +2339,8 @@ process_include_rules (MenuLayoutNode     *layout,
       {
         MenuLayoutNode *child;
 
+	menu_verbose ("Processing <And>\n");
+
         child = menu_layout_node_get_children (layout);
         while (child != NULL)
           {
@@ -2133,6 +2366,7 @@ process_include_rules (MenuLayoutNode     *layout,
 
             child = menu_layout_node_get_next (child);
           }
+	menu_verbose ("Processed <And>\n");
       }
       break;
 
@@ -2140,6 +2374,8 @@ process_include_rules (MenuLayoutNode     *layout,
       {
         MenuLayoutNode *child;
 
+	menu_verbose ("Processing <Or>\n");
+
         child = menu_layout_node_get_children (layout);
         while (child != NULL)
           {
@@ -2159,6 +2395,7 @@ process_include_rules (MenuLayoutNode     *layout,
 
             child = menu_layout_node_get_next (child);
           }
+	menu_verbose ("Processed <Or>\n");
       }
       break;
 
@@ -2167,6 +2404,8 @@ process_include_rules (MenuLayoutNode     *layout,
         /* First get the OR of all the rules */
         MenuLayoutNode *child;
 
+	menu_verbose ("Processing <Not>\n");
+
         child = menu_layout_node_get_children (layout);
         while (child != NULL)
           {
@@ -2192,18 +2431,24 @@ process_include_rules (MenuLayoutNode     *layout,
             /* Now invert the result */
             entry_directory_list_invert_set (list, set);
           }
+	menu_verbose ("Processed <Not>\n");
       }
       break;
 
     case MENU_LAYOUT_NODE_ALL:
+      menu_verbose ("Processing <All>\n");
       set = desktop_entry_set_new ();
       entry_directory_list_get_all_desktops (list, set);
+      menu_verbose ("Processed <All>\n");
       break;
 
     case MENU_LAYOUT_NODE_FILENAME:
       {
         DesktopEntry *entry;
 
+	menu_verbose ("Processing <Filename>%s</Filename>\n",
+		      menu_layout_node_get_content (layout));
+
         entry = entry_directory_list_get_desktop (list,
                                                   menu_layout_node_get_content (layout));
         if (entry != NULL)
@@ -2214,14 +2459,20 @@ process_include_rules (MenuLayoutNode     *layout,
                                          menu_layout_node_get_content (layout));
             desktop_entry_unref (entry);
           }
+	menu_verbose ("Processed <Filename>%s</Filename>\n",
+		      menu_layout_node_get_content (layout));
       }
       break;
 
     case MENU_LAYOUT_NODE_CATEGORY:
+      menu_verbose ("Processing <Category>%s</Category>\n",
+		    menu_layout_node_get_content (layout));
       set = desktop_entry_set_new ();
       entry_directory_list_get_by_category (list,
                                             menu_layout_node_get_content (layout),
                                             set);
+      menu_verbose ("Processed <Category>%s</Category>\n",
+		    menu_layout_node_get_content (layout));
       break;
 
     default:
@@ -2231,17 +2482,9 @@ process_include_rules (MenuLayoutNode     *layout,
   if (set == NULL)
     set = desktop_entry_set_new (); /* create an empty set */
 
-  return set;
-}
+  menu_verbose ("Matched %d entries\n", desktop_entry_set_get_count (set));
 
-static void
-mark_allocated_foreach (const char   *desktop_file_id,
-			DesktopEntry *desktop_entry,
-			GHashTable   *allocated)
-{
-  g_hash_table_insert (allocated,
-		       desktop_entry,
-		       desktop_entry_ref (desktop_entry));
+  return set;
 }
 
 static void
@@ -2261,15 +2504,15 @@ static MenuTreeDirectory *
 process_layout (MenuTree          *tree,
                 MenuTreeDirectory *parent,
                 MenuLayoutNode    *layout,
-                GHashTable        *allocated)
+                DesktopEntrySet   *allocated)
 {
   EntryDirectoryList *app_dirs;
   EntryDirectoryList *dir_dirs;
   MenuLayoutNode     *layout_iter;
   MenuTreeDirectory  *directory;
   DesktopEntrySet    *entries;
-  DesktopEntrySet    *allocated_set;
   gboolean            deleted;
+  DesktopEntrySet    *allocated_set;
   gboolean            only_unallocated;
   GSList             *tmp;
 
@@ -2301,6 +2544,8 @@ process_layout (MenuTree          *tree,
           {
             MenuTreeDirectory *child_dir;
 
+	    menu_verbose ("Processing <Menu>\n");
+
             child_dir = process_layout (tree,
                                         directory,
                                         layout_iter,
@@ -2308,6 +2553,8 @@ process_layout (MenuTree          *tree,
             if (child_dir)
               directory->subdirs = g_slist_prepend (directory->subdirs,
                                                     child_dir);
+
+	    menu_verbose ("Processed <Menu>\n");
           }
           break;
 
@@ -2319,6 +2566,9 @@ process_layout (MenuTree          *tree,
              */
             MenuLayoutNode *rule;
 
+	    menu_verbose ("Processing <Include> (%d entries)\n",
+			  desktop_entry_set_get_count (entries));
+
             rule = menu_layout_node_get_children (layout_iter);
             while (rule != NULL)
               {
@@ -2334,6 +2584,9 @@ process_layout (MenuTree          *tree,
 
                 rule = menu_layout_node_get_next (rule);
               }
+
+	    menu_verbose ("Processed <Include> (%d entries)\n",
+			  desktop_entry_set_get_count (entries));
           }
           break;
 
@@ -2345,6 +2598,9 @@ process_layout (MenuTree          *tree,
              */
             MenuLayoutNode *rule;
 
+	    menu_verbose ("Processing <Exclude> (%d entries)\n",
+			  desktop_entry_set_get_count (entries));
+
             rule = menu_layout_node_get_children (layout_iter);
             while (rule != NULL)
               {
@@ -2359,6 +2615,9 @@ process_layout (MenuTree          *tree,
 
                 rule = menu_layout_node_get_next (rule);
               }
+
+	    menu_verbose ("Processed <Exclude> (%d entries)\n",
+			  desktop_entry_set_get_count (entries));
           }
           break;
 
@@ -2366,7 +2625,10 @@ process_layout (MenuTree          *tree,
           {
             DesktopEntry *entry;
 
-            /*
+	    menu_verbose ("Processing <Directory>%s</Directory>\n",
+			  menu_layout_node_get_content (layout_iter));
+
+	    /*
              * The last <Directory> to exist wins, so we always try overwriting
              */
             entry = entry_directory_list_get_directory (dir_dirs,
@@ -2392,18 +2654,22 @@ process_layout (MenuTree          *tree,
           break;
 
         case MENU_LAYOUT_NODE_DELETED:
+	  menu_verbose ("Processed <Deleted/>\n");
           deleted = TRUE;
           break;
 
         case MENU_LAYOUT_NODE_NOT_DELETED:
+	  menu_verbose ("Processed <NotDeleted/>\n");
           deleted = FALSE;
           break;
 
         case MENU_LAYOUT_NODE_ONLY_UNALLOCATED:
+	  menu_verbose ("Processed <OnlyUnallocated/>\n");
           only_unallocated = TRUE;
           break;
 
         case MENU_LAYOUT_NODE_NOT_ONLY_UNALLOCATED:
+	  menu_verbose ("Processed <NotOnlyUnallocated/>\n");
           only_unallocated = FALSE;
           break;
 
@@ -2417,9 +2683,7 @@ process_layout (MenuTree          *tree,
   directory->only_unallocated = only_unallocated;
 
   if (!directory->only_unallocated)
-    desktop_entry_set_foreach (allocated_set,
-			       (DesktopEntrySetForeachFunc) mark_allocated_foreach,
-			       allocated);
+    desktop_entry_set_union (allocated, allocated_set);
 
   desktop_entry_set_unref (allocated_set);
 
@@ -2505,7 +2769,7 @@ process_layout (MenuTree          *tree,
 
 static void
 process_only_unallocated (MenuTreeDirectory *directory,
-                          GHashTable        *allocated)
+			  DesktopEntrySet   *allocated)
 {
   GSList *tmp;
 
@@ -2521,7 +2785,7 @@ process_only_unallocated (MenuTreeDirectory *directory,
           MenuTreeEntry *entry = tmp->data;
           GSList        *next  = tmp->next;
 
-          if (g_hash_table_lookup (allocated, entry->desktop_entry) != NULL)
+          if (desktop_entry_set_lookup (allocated, entry->desktop_file_id))
             {
               directory->entries = g_slist_delete_link (directory->entries,
                                                         tmp);
@@ -2571,7 +2835,7 @@ handle_entries_changed (MenuLayoutNode *layout,
 static void
 menu_tree_build_from_layout (MenuTree *tree)
 {
-  GHashTable *allocated;
+  DesktopEntrySet *allocated;
 
   if (tree->root)
     return;
@@ -2582,8 +2846,7 @@ menu_tree_build_from_layout (MenuTree *tree)
 
   menu_verbose ("Building menu tree from layout\n");
 
-  allocated = g_hash_table_new_full (NULL, NULL, NULL,
-				     (GDestroyNotify) desktop_entry_unref);
+  allocated = desktop_entry_set_new ();
 
   tree->root = process_layout (tree,
                                NULL,
@@ -2598,7 +2861,7 @@ menu_tree_build_from_layout (MenuTree *tree)
                                                  tree);
     }
 
-  g_hash_table_destroy (allocated);
+  desktop_entry_set_unref (allocated);
 }
 
 static void
